from typing import Callable, Type

from pydantic import BaseModel

from .exceptions import ExistingRecord, NonexistentRecord
from .results import PydandyResults
from .utils import PydandticProxy, PydanticData


class Table(PydanticData):
    pass


class PydandyTable(PydandticProxy):
    support_model = Table

    def __init__(
        self,
        model: Type[BaseModel],
        model_name: str,
        *args,
        **kwargs,
    ) -> None:
        self._model = model
        self._model_name = model_name
        super().__init__(*args, **kwargs)

    def __repr__(self) -> str:
        return f"PydandyTable<{self._model_name}>"

    def add(self, obj: BaseModel):
        assert isinstance(obj, self._model), f"Invalid type. Cannot add {obj.__class__} to {self}"
        record_id = hash(obj)
        if record_id in self:
            raise ExistingRecord
        self[record_id] = obj.copy()

    def get(self, id: int) -> BaseModel:
        if id not in self:
            raise NonexistentRecord
        return self[id].copy()

    def filter(self, filter: Callable) -> PydandyResults:
        return PydandyResults(records=[record.copy() for record in self.values() if filter(record)])

    def update(self, obj: BaseModel):
        assert isinstance(obj, self._model), f"Invalid type. Cannot add {obj.__class__} to {self}"
        record_id = hash(obj)
        if record_id not in self:
            raise NonexistentRecord
        self[record_id] = obj.copy()

    def delete(self, id: int):
        if id not in self:
            raise NonexistentRecord
        del self[id]



from enum import Enum, auto
from typing import Type

from pydantic import BaseModel

from .table import PydandyTable
from .utils import PydandticProxy, PydanticData


class DataMode(str, Enum):
    DIRECTORY = auto()
    FILE = auto()
    IN_MEMORY = auto()


class Data(PydanticData):
    pass


class PydandyData(PydandticProxy):
    support_model = Data

    def __init__(self) -> None:
        super().__init__()
        self._table_map: dict[str, PydandyTable] = {}

    def add_model(self, model: Type[BaseModel], name: str = None):
        model_name = name or model.__name__
        assert model_name not in self, f"A Model called {model_name} already exists!"
        new_table = PydandyTable(model, model_name)
        self._table_map[model_name] = new_table
        self[model_name] = new_table._data

    def get_table(self, name: str):
        return self._table_map.get(name)



from __future__ import annotations

from pathlib import Path
from typing import Hashable

from filelock import SoftFileLock
from pydantic import BaseModel

PydandyLock = SoftFileLock
PydandySource = Path | str | None

PydandyTable = dict[Hashable, BaseModel]



class ExistingRecord(Exception):
    def __init__(
        self,
        msg: str = "A record already exists with the given ID",
        *args: object,
    ) -> None:
        super().__init__(msg, *args)


class NonexistentRecord(Exception):
    def __init__(
        self,
        msg: str = "No record exists with the given ID",
        *args: object,
    ) -> None:
        super().__init__(msg, *args)



from pathlib import Path
from typing import Type, cast

from filelock import SoftFileLock
from pydantic import BaseModel

from .data import DataMode, PydandyData
from .types import PydandyLock, PydandySource


class PydandyDB:
    def __init__(
        self,
        data_source: PydandySource = None,
    ) -> None:
        self._data = PydandyData()

        self.data_source = Path(data_source) if data_source is not None else None

        self.mode = self._set_mode()
        if self.mode != DataMode.IN_MEMORY and self.data_source is not None:
            self._lock: PydandyLock = SoftFileLock(self.data_source / ".lock")
        else:
            # Fake Lock for consistent behavior/types
            self._lock: PydandyLock = SoftFileLock("/tmp/pydandy.lock")

    def __getattribute__(self, __name: str):
        _data = super().__getattribute__("_data") or {}
        if __name in _data:
            _data = cast(PydandyData, _data)
            return _data.get_table(__name)
        else:
            return super().__getattribute__(__name)

    def _set_mode(self) -> DataMode:
        if self.data_source is None:
            return DataMode.IN_MEMORY

        elif not self.data_source.exists():
            if self.data_source.suffix == ".json":
                self.data_source.parent.mkdir(parents=True, exist_ok=True)
                with open(self.data_source, "w+") as f:
                    f.write(self._data.json(sort_keys=True, indent=4))
                return DataMode.FILE

            if self.data_source.suffix == "":
                self.data_source.mkdir(parents=True, exist_ok=True)
                return DataMode.DIRECTORY

        elif self.data_source.is_file() and self.data_source.suffix == ".json":
            return DataMode.FILE

        elif self.data_source.is_dir():
            return DataMode.DIRECTORY

        raise ValueError("Invalid data source. Must be a directory, JSON file, or None")

    def add_model(self, model: Type[BaseModel], name: str = None) -> None:
        self._data.add_model(model, name)

    def register(self, name: str = None):
        def _register(model: Type[BaseModel]):
            self._data.add_model(model, name)
            return model

        return _register



from typing import Type

from pydantic import BaseModel


class PydanticData(BaseModel):
    __root__: dict[str | int, BaseModel] = {}


class PydandticProxy:
    support_model: Type[PydanticData] = PydanticData

    def __init__(self) -> None:
        self._data = self.support_model()

    def json(self, *args, **kwargs):
        return self._data.json(*args, **kwargs)

    def keys(self):
        return self._data.__root__.keys()

    def values(self):
        return self._data.__root__.values()

    def items(self):
        return self._data.__root__.items()

    def __iter__(self):
        return iter(self._data.__root__)

    def __getitem__(self, key: str | int):
        return self._data.__root__[key]

    def __setitem__(self, key: str | int, value: BaseModel):
        self._data.__root__[key] = value

    def __delitem__(self, key: str | int):
        del self._data.__root__[key]

    def __len__(self):
        return len(self._data.__root__)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pydandy']

package_data = \
{'': ['*']}

install_requires = \
['filelock>=3.7.1,<4.0.0', 'pydantic>=1.9.1,<2.0.0']

setup_kwargs = {
    'name': 'pydandy',
    'version': '0.0.1',
    'description': 'A handy-dandy datastore built on top of Pydantic',
    'long_description': '# Pydandy\n\nPydandtic + Handy\n\nA lightweight "Database", built on top of [Pydandtic](https://pydantic-docs.helpmanual.io/). It is currently under development, but the intent is to support in-memory, file, and directory storage options and eventually an Async variant. For the file and directory modes, the data will be backed by either one or manyu JSON files. The main database is also meant to provide a slim ORM-like interface for querying the data.\n\n## Examples\n```python\nfrom pydantic import BaseModel\nfrom pydandy import PydandyDB\n\n# Create an in-memory database\ndb = PydandyDB()\n# Add User Model to the database\n\n@db.register()\nclass User(BaseModel):\n    id: int\n    first_name: str\n    last_name: str\n    # You can use any model, as long you provide a __hash__\n    def __hash__(self):\n        return self.id\n\n# Add a new Record to Users\ndb.User.add(\n    User(\n        id=1,\n        first_name="John",\n        last_name="Baz",\n    )\n)\n\n# Get your record back\nuser = db.User.get(1)\n\n# Filter for records\ndb.User.filter(lambda record: record.first_name.startswith("J"))\n```\n\n## Motivation\nMostly just because, but also because I occasionaly finding myself wanting a small data, portable data store. This seemed like fun project idea, and I really Pydandtic, so it worked out.\n\n## Contributing\nAt this stage, I am not accepting contributions. Mostly because I am still trying to shape out the core functionality. However, this should change soon™ if you are interested in helping out.',
    'author': 'Will Johns',
    'author_email': 'will@wcj.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/WilliamJohns/pydandy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
